/* ========================================================================

    KTOPS v0.89f - a simple TopTen-bulletin generator for use with BBBS.
    Copyright (C) 1994 Hannu Laurila (bugfix by Mr. B (B as bug))

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


    Contacting the author:  Hannu Laurila
                            Vrmentie 131
                            FIN-62900 ALAJRVI

                            EMail...: Hancat@freeport.uwasa.fi
                            Fidonet.: Hannu Laurila @ 2:224/966.0
                            BBS.....: +358-66-5574558

    Note that file bbbsdef.h included in distribution archive ktop_89u.lzh
    is *not* distributed under the terms of GNU General Public License.

    -----------------------------------------------------------------------

     * KTOPS is copyright 1994, Hannu Laurila (bugfix by Mr.B (B as bug))
     * BBBS is copyright 1990, 1992-1994, Kim Heino and Tapani T. Salmi

    Please post modified/enchanced versions to me!

   ------------------------------------------------------------------------

    USAGE: KTOPS [usrfile] [bulletinfile]

        [usrfile]       = path to your bbbsusr.dat file
        [bulletinfile]  = path to your TopTen-bulletin

    Note! In this form program is not fully file-sharing aware, please
    be careful in multitasking environments.

   ======================================================================== */

#if defined(__OS2__) /* bbbsdef.h requires os2.h if defined __OS2__ */
 #include <os2.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "bbbsdef.h"    /* bbbsdef.h from BBBS package */

#define VERNAME "KTOPS v0.89u"

struct toprec {
  char        name[30];
  word        points;
};

struct toprec       top[4][10]={0};
struct userrec      u;
FILE                *userf, *bullf;    /* bbbsusr.dat & bulletin file */

/* -- prototypes ---------------------------------------------------- */
void checktop (byte topnumber, word vrt);
void makelist (byte listnumber);


int main (int argc, char **argv)

{
  ulong totalusers,totalmsgr,totalmsgd;
  ulong msgleft,logons,uploads,downloads;

  char time_str[25]; struct tm *tp; time_t ltime;

  totalusers=totalmsgr=totalmsgd=msgleft=logons=uploads=downloads=0;

  if (argc!=3 && argc!=2) {
    fprintf(stderr,"Usage: %s [userfile] [bulletinfile]\n",argv[0]);
    fprintf(stderr,"if [bulletinfile] is not specified, stdout will be used instead\n");
    return 1;
  }

  if ( (userf=fopen(argv[1],"rb")) == NULL ) {
    fprintf(stderr,"Couldn't open file %s for binary read!",argv[1]);
    return 2;
  }

  fprintf(stderr,"%s - (c) 1994 Hannu Laurila.\n",VERNAME);

  while (fread(&u, sizeof(u), 1, userf))
  {
    totalusers++;
    totalmsgr+=(ulong)u.msgread; totalmsgd+=(ulong)u.msgdumped;
    msgleft+=(ulong)u.msgleft;  logons+=(ulong)u.timeson;
    uploads+=(ulong)u.uploaded; downloads+=(ulong)u.downloaded;

    if (u.ok2login!=2) {
      checktop(0, u.timeson);     checktop(1, u.msgleft);
      checktop(2, u.uploaded);    checktop(3, u.downloaded);
    }
  }
  fclose(userf);

  if (argc==2) bullf=stdout;
   else
  if ( (bullf=fopen(argv[2],"wt")) == NULL ) {
    fprintf(stderr,"Couldn't open file %s for text write!",argv[2]);
    exit (2);
  }

  time(&ltime); tp = localtime (&ltime);
  strftime (time_str,sizeof(time_str),"%a %d.%m.%Y, %H:%M.",tp);

  fprintf(bullf,"\n     TopTen-lists, created with %s",VERNAME);
  fprintf(bullf,"\n     Last updated %s",time_str);

  fprintf(bullf,"\n\n     Total users on system:          %lu",totalusers);
  fprintf(bullf,"\n     Total messages read by users:   %lu",totalmsgr);
  fprintf(bullf,"\n     Total messages dumped by users: %lu\n",totalmsgd);

  fprintf(bullf,"\n\n     Most logons           (total: %7lu)\n\n",logons);
  makelist(0);
  fprintf(bullf,"\n\n     Most messages left    (total: %7lu)\n\n",msgleft);
  makelist(1);
  fprintf(bullf,"\n\n     Most uploads          (total: %7lu)\n\n",uploads);
  makelist(2);
  fprintf(bullf,"\n\n     Most downloads        (total: %7lu)\n\n",downloads);
  makelist(3);
  fprintf(bullf,"\n     %s, (c) 1994 Hannu Laurila\n\n",VERNAME);
  fclose(bullf);
  return 0;
}

void checktop (register byte topnumber, register word vrt)

{
  register byte foo1,foo2;
  for (foo1=0;foo1<10;foo1++) {
    if (top[topnumber][foo1].points < vrt) {
      for (foo2=9; foo2>foo1; foo2--)
        memcpy(&top[topnumber][foo2],&top[topnumber][foo2-1],sizeof(struct toprec));
      strcpy(top[topnumber][foo1].name,u.name);
      top[topnumber][foo1].points=vrt;
      return;
    } /* if */
  } /* for */
  return;
}

void makelist (byte listnumber)

#define TOPFORMAT   " %2u. %-30s%7u\n"
#define C_TOPFORMAT " %2u. \033[%sm%-30s\033[0m%7u\n"

{
  byte foo1;
  for (foo1=0;foo1<10;foo1++) {
    if (foo1>2) fprintf(bullf,TOPFORMAT,foo1+1,top[listnumber][foo1].name,top[listnumber][foo1].points);
   else
    fprintf(bullf,C_TOPFORMAT,foo1+1,
       foo1==0 ? "0;1" : "0;32", top[listnumber][foo1].name,top[listnumber][foo1].points);
  }
}
