/* ftell.c (emx+gcc) -- Copyright (c) 1990-1995 by Eberhard Mattes */

#include <sys/emx.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>

long ftell (FILE *stream)
{
  long pos;

  if (stream->flags & _IOSPECIAL)
    pos = 0;
  else
    {
      pos = tell (stream->handle);
      if (pos == -1)
        return (-1);
    }
  if (stream->flags & _IOWRT)
    {
      if (bbuf (stream))
        {
          pos += stream->ptr - stream->buffer;
          if (!(stream->flags & _IOSPECIAL)
              && fileno (stream) >= 0 && fileno (stream) < _nfiles
              && _files[fileno (stream)] & O_TEXT)
            {
              const char *p;
              int n;

              /* In text mode, newlines are translated to CR/LF pairs.
                 Adjust the position to take account of this. */

              n = stream->ptr - stream->buffer;
              p = stream->buffer;
              while (n > 0)
                {
                  if (*p ++ == '\n')
                    ++pos;
                  --n;
                }
            }
        }
    }
  else if (stream->flags & _IOREAD)
    {
      /* Subtract the number of unprocessed buffered characters from
         the file position of the end of the buffer to get the current
         position. */

      if (!(stream->flags & _IOSPECIAL)
          && fileno (stream) >= 0 && fileno (stream) < _nfiles
          && (_files[fileno (stream)] & F_CRLF)
          && stream->_ungetc_count == 0)
        {
          const char *p;
          int n;

          /* There has been at least one CR/LF to newline translation.
             Assume that all newlines in the buffer have been
             translated from a CR/LF pair each. */

          p = stream->ptr;
          n = stream->rcount;
          pos -= n;
          while (n > 0)
            {
              if (*p++ == '\n')
                --pos;
              --n;
            }
        }
      else if (stream->_ungetc_count != 0)
        {
          /* There are characters pushed back by ungetc().  For binary
             streams, each successful ungetc() decrements the file position
             indicator.  Treat text mode like binary mode as the file
             position indicator is undefined in text mode after
             calling ungetc(). */

          pos = pos + stream->rcount - stream->_ungetc_count;
        }
      else
        {
          /* Binary mode. */

          pos -= stream->rcount;
        }
    }
  return (pos);
}
