/* stdiotst.c (emx+gcc) */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* Test ftell(). */

static void test_t (const char *fname)
{
  FILE *f;
  int i, j;

  f = fopen (fname, "wb");
  if (f == NULL)
    {
      perror ("fopen");
      exit (1);
    }
  for (i = 0; i < 32768; ++i)
    {
      assert (fwrite (&i, sizeof (i), 1, f) == 1);
      if (i % 4111 == 0)
        assert (fflush (f) == 0);
      assert (ftell (f) == (i+1) * sizeof (i));
    }
  assert (fflush (f) == 0);
  assert (fclose (f) == 0);
  f = fopen (fname, "rb");
  if (f == NULL)
    {
      perror ("fopen");
      exit (1);
    }
  for (i = 0; i < 32768; ++i)
    {
      assert (fread (&j, sizeof (j), 1, f) == 1);
      assert (i == j);
      if (i % 4111 == 0)
        assert (fflush (f) == 0);
      assert (ftell (f) == (i+1) * sizeof (i));
    }
  assert (fclose (f) == 0);
}


/* Test switching between `read mode' and `write mode' of a stream
   open for update. */

static void test_u (void)
{
  FILE *f;
  char buf[10];

  f = tmpfile ();
  if (f == NULL)
    {
      perror ("tmpfile()");
      exit (2);
    }
  assert (fwrite ("abc", 1, 3, f) == 3);
  rewind (f);
  assert (fread (buf, 1, 10, f) == 3);
  assert (memcmp (buf, "abc", 3) == 0);
  assert (fseek (f, 0L, SEEK_SET) == 0);
  assert (fputc ('A', f) == 'A');
  assert (fseek (f, 0L, SEEK_SET) == 0);
  assert (fread (buf, 1, 2, f) == 2);
  assert (memcmp (buf, "Ab", 2) == 0);
  assert (fseek (f, 0L, SEEK_CUR) == 0);
  assert (fputc ('C', f) == 'C');
  assert (fseek (f, 0L, SEEK_SET) == 0);
  assert (fread (buf, 1, 10, f) == 3);
  assert (memcmp (buf, "AbC", 3) == 0);
  /* At EOF, we don't have to switch explicitely to write mode. */
  assert (fputc ('D', f) == 'D');
  assert (fseek (f, 0L, SEEK_SET) == 0);
  assert (fread (buf, 1, 10, f) == 4);
  assert (memcmp (buf, "AbCD", 4) == 0);

  assert (fseek (f, 1L, SEEK_SET) == 0);
  assert (fputc ('B', f) == 'B');
  assert (fflush (f) == 0);
  assert (fgetc (f) == 'C');
  assert (fseek (f, 0L, SEEK_SET) == 0);
  assert (fread (buf, 1, 10, f) == 4);
  assert (memcmp (buf, "ABCD", 4) == 0);

  assert (fseek (f, 3L, SEEK_SET) == 0);
  assert (fgetc (f) == 'D');
  assert (fgetc (f) == EOF);
  assert (fputc ('E', f) == 'E');
  assert (fseek (f, 0L, SEEK_SET) == 0);
  assert (fread (buf, 1, 10, f) == 5);
  assert (memcmp (buf, "ABCDE", 5) == 0);

  assert (fclose (f) == 0);
}


int main (int argc, char *argv[])
{

  if (argc == 3 && strcmp (argv[1], "-t") == 0)
    test_t (argv[2]);
  else if (argc == 2 && strcmp (argv[1], "-u") == 0)
    test_u ();
  else
    {
      fputs ("Usage: stdiotst -t <output_file>\n", stderr);
      fputs ("       stdiotst -u\n", stderr);
      exit (1);
    }
  return (0);
}
